@@ -1,7 +1,7 @@
 /*
  * server.c - Provide shadowsocks service
  *
- * Copyright (C) 2013 - 2018, Max Lv <max.c.lv@gmail.com>
+ * Copyright (C) 2013 - 2019, Max Lv <max.c.lv@gmail.com>
  *
  * This file is part of the shadowsocks-libev.
  *
@@ -212,8 +212,7 @@ construct_command_line(struct manager_ctx *manager, struct server *server)
         int len = strlen(cmd);
         snprintf(cmd + len, BUF_SIZE - len, " -d \"%s\"", manager->nameservers);
     }
-    if (manager->workdir)
-    {
+    if (manager->workdir) {
         int len = strlen(cmd);
         snprintf(cmd + len, BUF_SIZE - len, " -D \"%s\"", manager->workdir);
     }
@@ -673,7 +672,7 @@ manager_recv_cb(EV_P_ ev_io *w, int revents)
         }
 
         size_t pos = strlen(buf);
-        strcpy(buf + pos - 1, "\n]"); // Remove trailing ","
+        strcpy(buf + max(pos - 1, 1), "\n]"); // Remove trailing ","
         pos = strlen(buf);
         if (sendto(manager->fd, buf, pos, 0, (struct sockaddr *)&claddr, len)
             != pos) {
@@ -709,6 +708,7 @@ manager_recv_cb(EV_P_ ev_io *w, int revents)
         }
 
         update_stat(port, traffic);
+
     } else if (strcmp(action, "ping") == 0) {
         struct cork_hash_table_entry *entry;
         struct cork_hash_table_iterator server_iter;
@@ -896,7 +896,7 @@ main(int argc, char **argv)
         { "password",        required_argument, NULL, GETOPT_VAL_PASSWORD    },
         { "workdir",         required_argument, NULL, GETOPT_VAL_WORKDIR     },
         { "help",            no_argument,       NULL, GETOPT_VAL_HELP        },
-        { NULL,                              0, NULL,                      0 }
+        { NULL,              0,                 NULL, 0                      }
     };
 
     opterr = 0;
@@ -1050,8 +1050,7 @@ main(int argc, char **argv)
         if (ipv6first == 0) {
             ipv6first = conf->ipv6_first;
         }
-        if (workdir == NULL)
-        {
+        if (workdir == NULL) {
             workdir = conf->workdir;
         }
         if (acl == NULL) {
@@ -1156,7 +1155,7 @@ main(int argc, char **argv)
     }
 #endif
 
-    struct passwd *pw   = getpwuid(getuid());
+    struct passwd *pw = getpwuid(getuid());
 
     if (workdir == NULL || strlen(workdir) == 0) {
         workdir = pw->pw_dir;
@@ -1166,11 +1165,11 @@ main(int argc, char **argv)
         }
 
         working_dir_size = strlen(workdir) + 15;
-        working_dir = ss_malloc(working_dir_size);
+        working_dir      = ss_malloc(working_dir_size);
         snprintf(working_dir, working_dir_size, "%s/.shadowsocks", workdir);
     } else {
         working_dir_size = strlen(workdir) + 2;
-        working_dir = ss_malloc(working_dir_size);
+        working_dir      = ss_malloc(working_dir_size);
         snprintf(working_dir, working_dir_size, "%s", workdir);
     }
     LOGI("working directory points to %s", working_dir);
